@@ -36,7 +36,7 @@ use winapi::{um::handleapi::INVALID_HANDLE_VALUE, um::winsock2::INVALID_SOCKET};
 // because c_int is 32 bits.
 #[cfg_attr(rustc_attrs, rustc_layout_scalar_valid_range_end(0xFF_FF_FF_FE))]
 pub struct BorrowedFd<'fd> {
-    raw: RawFd,
+    fd: RawFd,
     _phantom: PhantomData<&'fd OwnedFd>,
 }
 
@@ -57,7 +57,7 @@ pub struct BorrowedFd<'fd> {
 #[derive(Copy, Clone)]
 #[repr(transparent)]
 pub struct BorrowedHandle<'handle> {
-    raw: NonNull<c_void>,
+    handle: NonNull<c_void>,
     _phantom: PhantomData<&'handle OwnedHandle>,
 }
 
@@ -84,7 +84,7 @@ pub struct BorrowedHandle<'handle> {
     rustc_layout_scalar_valid_range_end(0xFF_FF_FF_FF_FF_FF_FF_FE)
 )]
 pub struct BorrowedSocket<'socket> {
-    raw: RawSocket,
+    socket: RawSocket,
     _phantom: PhantomData<&'socket OwnedSocket>,
 }
 
@@ -104,7 +104,7 @@ pub struct BorrowedSocket<'socket> {
 // because c_int is 32 bits.
 #[cfg_attr(rustc_attrs, rustc_layout_scalar_valid_range_end(0xFF_FF_FF_FE))]
 pub struct OwnedFd {
-    raw: RawFd,
+    fd: RawFd,
 }
 
 /// An owned handle.
@@ -117,14 +117,14 @@ pub struct OwnedFd {
 ///
 /// Note that it *may* have the value [`INVALID_HANDLE_VALUE`]. See [here] for
 /// the full story. For APIs like `CreateFileW` which report errors with
-/// `INVALID_HANDLE_VALUE` instead of null, use [`OptionFileHandle`] instead
+/// `INVALID_HANDLE_VALUE` instead of null, use [`HandleOrInvalid`] instead
 /// of `Option<OwnedHandle>`.
 ///
 /// [here]: https://devblogs.microsoft.com/oldnewthing/20040302-00/?p=40443
 #[cfg(windows)]
 #[repr(transparent)]
 pub struct OwnedHandle {
-    raw: NonNull<c_void>,
+    handle: NonNull<c_void>,
 }
 
 /// An owned socket.
@@ -148,24 +148,24 @@ pub struct OwnedHandle {
     rustc_layout_scalar_valid_range_end(0xFF_FF_FF_FF_FF_FF_FF_FE)
 )]
 pub struct OwnedSocket {
-    raw: RawSocket,
+    socket: RawSocket,
 }
 
-/// Similar to `Option<OwnedHandle>`, but intended for use in FFI interfaces
-/// where [`INVALID_HANDLE_VALUE`] is used as the sentry value, and null values
-/// are not used at all, such as in the return value of `CreateFileW`.
+/// FFI type for handles in return values or out parameters, where `INVALID_HANDLE_VALUE` is used
+/// as a sentry value to indicate errors, such as in the return value of `CreateFileW`. This uses
+/// `repr(transparent)` and has the representation of a host handle, so that it can be used in such
+/// FFI declarations.
 ///
-/// If this holds an owned handle, it closes the handle on drop.
+/// The only thing you can usefully do with a `HandleOrInvalid` is to convert it into an
+/// `OwnedHandle` using its [`TryFrom`] implementation; this conversion takes care of the check for
+/// `INVALID_HANDLE_VALUE`. This ensures that such FFI calls cannot start using the handle without
+/// checking for `INVALID_HANDLE_VALUE` first.
 ///
-/// This uses `repr(transparent)` and has the representation of a host handle,
-/// so it can be used in FFI in places where a non-null handle is passed as a
-/// consumed argument or returned as an owned value, or it is
-/// [`INVALID_HANDLE_VALUE`] indicating an error or an otherwise absent value.
+/// If this holds a valid handle, it will close the handle on drop.
 #[cfg(windows)]
 #[repr(transparent)]
-pub struct OptionFileHandle {
-    raw: RawHandle,
-}
+#[derive(Debug)]
+pub struct HandleOrInvalid(Option<OwnedHandle>);
 
 // The Windows [`HANDLE`] type may be transferred across and shared between
 // thread boundaries (despite containing a `*mut void`, which in general isn't
@@ -175,13 +175,13 @@ pub struct OptionFileHandle {
 #[cfg(windows)]
 unsafe impl Send for OwnedHandle {}
 #[cfg(windows)]
-unsafe impl Send for OptionFileHandle {}
+unsafe impl Send for HandleOrInvalid {}
 #[cfg(windows)]
 unsafe impl Send for BorrowedHandle<'_> {}
 #[cfg(windows)]
 unsafe impl Sync for OwnedHandle {}
 #[cfg(windows)]
-unsafe impl Sync for OptionFileHandle {}
+unsafe impl Sync for HandleOrInvalid {}
 #[cfg(windows)]
 unsafe impl Sync for BorrowedHandle<'_> {}
 
@@ -194,10 +194,10 @@ impl BorrowedFd<'_> {
     /// The resource pointed to by `raw` must remain open for the duration of
     /// the returned `BorrowedFd`, and it must not have the value `-1`.
     #[inline]
-    pub unsafe fn borrow_raw_fd(raw: RawFd) -> Self {
-        debug_assert_ne!(raw, -1_i32 as RawFd);
+    pub unsafe fn borrow_raw_fd(fd: RawFd) -> Self {
+        debug_assert_ne!(fd, -1_i32 as RawFd);
         Self {
-            raw,
+            fd,
             _phantom: PhantomData,
         }
     }
@@ -212,10 +212,10 @@ impl BorrowedHandle<'_> {
     /// The resource pointed to by `raw` must remain open for the duration of
     /// the returned `BorrowedHandle`, and it must not be null.
     #[inline]
-    pub unsafe fn borrow_raw_handle(raw: RawHandle) -> Self {
-        debug_assert!(!raw.is_null());
+    pub unsafe fn borrow_raw_handle(handle: RawHandle) -> Self {
+        debug_assert!(!handle.is_null());
         Self {
-            raw: NonNull::new_unchecked(raw),
+            handle: NonNull::new_unchecked(handle),
             _phantom: PhantomData,
         }
     }
@@ -231,72 +231,46 @@ impl BorrowedSocket<'_> {
     /// the returned `BorrowedSocket`, and it must not have the value
     /// [`INVALID_SOCKET`].
     #[inline]
-    pub unsafe fn borrow_raw_socket(raw: RawSocket) -> Self {
-        debug_assert_ne!(raw, INVALID_SOCKET as RawSocket);
+    pub unsafe fn borrow_raw_socket(socket: RawSocket) -> Self {
+        debug_assert_ne!(socket, INVALID_SOCKET as RawSocket);
         Self {
-            raw,
+            socket,
             _phantom: PhantomData,
         }
     }
 }
 
 #[cfg(windows)]
-impl OptionFileHandle {
-    /// Return an empty `OptionFileHandle` with no resource.
-    #[inline]
-    pub const fn none() -> Self {
-        Self {
-            raw: INVALID_HANDLE_VALUE,
-        }
-    }
-}
-
-#[cfg(windows)]
-impl TryFrom<OptionFileHandle> for OwnedHandle {
+impl TryFrom<HandleOrInvalid> for OwnedHandle {
     type Error = ();
 
     #[inline]
-    fn try_from(option: OptionFileHandle) -> Result<Self, ()> {
-        let raw = option.raw;
-        forget(option);
-        if let Some(non_null) = NonNull::new(raw) {
-            if non_null.as_ptr() != INVALID_HANDLE_VALUE {
-                Ok(Self { raw: non_null })
-            } else {
-                Err(())
-            }
+    fn try_from(handle_or_invalid: HandleOrInvalid) -> Result<Self, ()> {
+        // In theory, we ought to be able to assume that the pointer here is
+        // never null, use `OwnedHandle` rather than `Option<OwnedHandle>`, and
+        // obviate the the panic path here.  Unfortunately, Win32 documentation
+        // doesn't explicitly guarantee this anywhere.
+        //
+        // APIs like [`CreateFileW`] itself have `HANDLE` arguments where a
+        // null handle indicates an absent value, which wouldn't work if null
+        // were a valid handle value, so it seems very unlikely that it could
+        // ever return null. But who knows?
+        //
+        // [`CreateFileW`]: https://docs.microsoft.com/en-us/windows/win32/api/fileapi/nf-fileapi-createfilew
+        let owned_handle = handle_or_invalid.0.expect("A `HandleOrInvalid` was null!");
+        if owned_handle.handle.as_ptr() == INVALID_HANDLE_VALUE {
+            Err(())
         } else {
-            // In theory, we ought to be able to assume that the pointer here
-            // is never null, change `option.raw` to `NonNull`, and obviate the
-            // the panic path here. Unfortunately, Win32 documentation doesn't
-            // explicitly guarantee this anywhere.
-            //
-            // APIs like [`CreateFileW`] itself have `HANDLE` arguments where a
-            // null handle indicates an absent value, which wouldn't work if
-            // null were a valid handle value, so it seems very unlikely that
-            // it could ever return null. But who knows?
-            //
-            // [`CreateFileW`]: https://docs.microsoft.com/en-us/windows/win32/api/fileapi/nf-fileapi-createfilew
-            panic!("An `OptionFileHandle` was null!");
+            Ok(owned_handle)
         }
     }
 }
 
-#[cfg(windows)]
-impl From<OwnedHandle> for OptionFileHandle {
-    #[inline]
-    fn from(owned: OwnedHandle) -> Self {
-        let raw = owned.raw;
-        forget(owned);
-        Self { raw: raw.as_ptr() }
-    }
-}
-
 #[cfg(any(unix, target_os = "wasi"))]
 impl AsRawFd for BorrowedFd<'_> {
     #[inline]
     fn as_raw_fd(&self) -> RawFd {
-        self.raw
+        self.fd
     }
 }
 
@@ -304,7 +278,7 @@ impl AsRawFd for BorrowedFd<'_> {
 impl AsRawHandle for BorrowedHandle<'_> {
     #[inline]
     fn as_raw_handle(&self) -> RawHandle {
-        self.raw.as_ptr()
+        self.handle.as_ptr()
     }
 }
 
@@ -312,7 +286,7 @@ impl AsRawHandle for BorrowedHandle<'_> {
 impl AsRawSocket for BorrowedSocket<'_> {
     #[inline]
     fn as_raw_socket(&self) -> RawSocket {
-        self.raw
+        self.socket
     }
 }
 
@@ -320,7 +294,7 @@ impl AsRawSocket for BorrowedSocket<'_> {
 impl AsRawFd for OwnedFd {
     #[inline]
     fn as_raw_fd(&self) -> RawFd {
-        self.raw
+        self.fd
     }
 }
 
@@ -328,7 +302,7 @@ impl AsRawFd for OwnedFd {
 impl AsRawHandle for OwnedHandle {
     #[inline]
     fn as_raw_handle(&self) -> RawHandle {
-        self.raw.as_ptr()
+        self.handle.as_ptr()
     }
 }
 
@@ -336,7 +310,7 @@ impl AsRawHandle for OwnedHandle {
 impl AsRawSocket for OwnedSocket {
     #[inline]
     fn as_raw_socket(&self) -> RawSocket {
-        self.raw
+        self.socket
     }
 }
 
@@ -344,9 +318,9 @@ impl AsRawSocket for OwnedSocket {
 impl IntoRawFd for OwnedFd {
     #[inline]
     fn into_raw_fd(self) -> RawFd {
-        let raw = self.raw;
+        let fd = self.fd;
         forget(self);
-        raw
+        fd
     }
 }
 
@@ -354,9 +328,9 @@ impl IntoRawFd for OwnedFd {
 impl IntoRawHandle for OwnedHandle {
     #[inline]
     fn into_raw_handle(self) -> RawHandle {
-        let raw = self.raw.as_ptr();
+        let handle = self.handle.as_ptr();
         forget(self);
-        raw
+        handle
     }
 }
 
@@ -364,9 +338,9 @@ impl IntoRawHandle for OwnedHandle {
 impl IntoRawSocket for OwnedSocket {
     #[inline]
     fn into_raw_socket(self) -> RawSocket {
-        let raw = self.raw;
+        let socket = self.socket;
         forget(self);
-        raw
+        socket
     }
 }
 
@@ -379,9 +353,9 @@ impl FromRawFd for OwnedFd {
     /// The resource pointed to by `raw` must be open and suitable for assuming
     /// ownership.
     #[inline]
-    unsafe fn from_raw_fd(raw: RawFd) -> Self {
-        debug_assert_ne!(raw, -1_i32 as RawFd);
-        Self { raw }
+    unsafe fn from_raw_fd(fd: RawFd) -> Self {
+        debug_assert_ne!(fd, -1_i32 as RawFd);
+        Self { fd }
     }
 }
 
@@ -394,10 +368,10 @@ impl FromRawHandle for OwnedHandle {
     /// The resource pointed to by `raw` must be open and suitable for assuming
     /// ownership.
     #[inline]
-    unsafe fn from_raw_handle(raw: RawHandle) -> Self {
-        debug_assert!(!raw.is_null());
+    unsafe fn from_raw_handle(handle: RawHandle) -> Self {
+        debug_assert!(!handle.is_null());
         Self {
-            raw: NonNull::new_unchecked(raw),
+            handle: NonNull::new_unchecked(handle),
         }
     }
 }
@@ -411,15 +385,20 @@ impl FromRawSocket for OwnedSocket {
     /// The resource pointed to by `raw` must be open and suitable for assuming
     /// ownership.
     #[inline]
-    unsafe fn from_raw_socket(raw: RawSocket) -> Self {
-        debug_assert_ne!(raw, INVALID_SOCKET as RawSocket);
-        Self { raw }
+    unsafe fn from_raw_socket(socket: RawSocket) -> Self {
+        debug_assert_ne!(socket, INVALID_SOCKET as RawSocket);
+        Self { socket }
     }
 }
 
 #[cfg(windows)]
-impl FromRawHandle for OptionFileHandle {
-    /// Constructs a new instance of `Self` from the given raw handle.
+impl FromRawHandle for HandleOrInvalid {
+    /// Constructs a new instance of `Self` from the given `RawHandle` returned
+    /// from a Windows API that uses `INVALID_HANDLE_VALUE` to indicate
+    /// failure, such as `CreateFileW`.
+    ///
+    /// Use `Option<OwnedHandle>` instead of `HandleOrInvalid` for APIs that
+    /// use null to indicate failure.
     ///
     /// # Safety
     ///
@@ -430,9 +409,9 @@ impl FromRawHandle for OptionFileHandle {
     ///
     /// [here]: https://devblogs.microsoft.com/oldnewthing/20040302-00/?p=40443
     #[inline]
-    unsafe fn from_raw_handle(raw: RawHandle) -> Self {
-        debug_assert!(!raw.is_null());
-        Self { raw }
+    unsafe fn from_raw_handle(handle: RawHandle) -> Self {
+        // We require non-null here to catch errors earlier.
+        Self(Some(OwnedHandle::from_raw_handle(handle)))
     }
 }
 
@@ -442,7 +421,7 @@ impl Drop for OwnedFd {
     fn drop(&mut self) {
         #[cfg(feature = "close")]
         unsafe {
-            let _ = libc::close(self.raw as std::os::raw::c_int);
+            let _ = libc::close(self.fd as std::os::raw::c_int);
         }
 
         // If the `close` feature is disabled, we expect users to avoid letting
@@ -462,7 +441,7 @@ impl Drop for OwnedHandle {
     #[inline]
     fn drop(&mut self) {
         unsafe {
-            let _ = winapi::um::handleapi::CloseHandle(self.raw.as_ptr());
+            let _ = winapi::um::handleapi::CloseHandle(self.handle.as_ptr());
         }
     }
 }
@@ -472,17 +451,7 @@ impl Drop for OwnedSocket {
     #[inline]
     fn drop(&mut self) {
         unsafe {
-            let _ = winapi::um::winsock2::closesocket(self.raw as winapi::um::winsock2::SOCKET);
-        }
-    }
-}
-
-#[cfg(windows)]
-impl Drop for OptionFileHandle {
-    #[inline]
-    fn drop(&mut self) {
-        unsafe {
-            let _ = winapi::um::handleapi::CloseHandle(self.raw);
+            let _ = winapi::um::winsock2::closesocket(self.socket as winapi::um::winsock2::SOCKET);
         }
     }
 }
@@ -491,7 +460,7 @@ impl Drop for OptionFileHandle {
 impl fmt::Debug for BorrowedFd<'_> {
     #[allow(clippy::missing_inline_in_public_items)]
     fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
-        f.debug_struct("BorrowedFd").field("fd", &self.raw).finish()
+        f.debug_struct("BorrowedFd").field("fd", &self.fd).finish()
     }
 }
 
@@ -500,7 +469,7 @@ impl fmt::Debug for BorrowedHandle<'_> {
     #[allow(clippy::missing_inline_in_public_items)]
     fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
         f.debug_struct("BorrowedHandle")
-            .field("handle", &self.raw)
+            .field("handle", &self.handle)
             .finish()
     }
 }
@@ -510,7 +479,7 @@ impl fmt::Debug for BorrowedSocket<'_> {
     #[allow(clippy::missing_inline_in_public_items)]
     fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
         f.debug_struct("BorrowedSocket")
-            .field("socket", &self.raw)
+            .field("socket", &self.socket)
             .finish()
     }
 }
@@ -519,7 +488,7 @@ impl fmt::Debug for BorrowedSocket<'_> {
 impl fmt::Debug for OwnedFd {
     #[allow(clippy::missing_inline_in_public_items)]
     fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
-        f.debug_struct("OwnedFd").field("fd", &self.raw).finish()
+        f.debug_struct("OwnedFd").field("fd", &self.fd).finish()
     }
 }
 
@@ -528,7 +497,7 @@ impl fmt::Debug for OwnedHandle {
     #[allow(clippy::missing_inline_in_public_items)]
     fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
         f.debug_struct("OwnedHandle")
-            .field("handle", &self.raw)
+            .field("handle", &self.handle)
             .finish()
     }
 }
@@ -538,17 +507,7 @@ impl fmt::Debug for OwnedSocket {
     #[allow(clippy::missing_inline_in_public_items)]
     fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
         f.debug_struct("OwnedSocket")
-            .field("socket", &self.raw)
-            .finish()
-    }
-}
-
-#[cfg(windows)]
-impl fmt::Debug for OptionFileHandle {
-    #[allow(clippy::missing_inline_in_public_items)]
-    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
-        f.debug_struct("OptionFileHandle")
-            .field("handle", &self.raw)
+            .field("socket", &self.socket)
             .finish()
     }
 }
